{*******************************************************************}
{                                                                   }
{  WheelMs.pas, Copyright (c) Simon Page 1998                       }
{                                                                   }
{  Version....: 1.13                                                }
{                                                                   }
{  Purpose....: Component for 'wheeled' mouse support on            }
{               Windows 95, 98 & NT                                 }
{                                                                   }
{  Requires...: Delphi 2 or greater                                 }
{                                                                   }
{*******************************************************************}

{$I WheelMs.inc}

{$IFNDEF WIN32} 'This component requires Delphi 2 or higher !' {$ENDIF}

{ Revisions:

  v1.00, 24 Jan 98

    Initial release

  v1.10, 24 Feb 98

    Revised storage system for 'SpecialCtrls'. New system based on custom
    list type object using a compact system for storing records in single
    block of memory. This system chosen in preference to a TCollection
    derivative due to quirky update behaviour when using collections to
    store component references and using form inheritance.

    Bug fixed that stopped TWheelMouse working on an MDI child form under
    Windows 95.

  v1.11, 15 Mar 98

    Bug fixed in design time behaviour which would prevent a TWheelMouse
    component from being dropped onto a form that had already been set as
    an MDI child form on Windows 95.

  v1.12, 30 Jun 98

    Bug fixed that would cause an AV on Windows 95 when multiple MDI
    child windows each with TWheelMouse components on were not freed in
    reverse order to creation!  Problem due to a 'broken' chain of
    window proc subclasses.

  v1.13, 14 Aug 98

   Delphi 4 compatibility update and a couple of minor internal changes.
}

unit WheelMs;

interface

uses
  Windows, Messages, SysUtils, Classes, Controls, Forms;

type

{$IFDEF VERSION3_DOWN}
  TWMMouseWheel = packed record
    Msg: Cardinal;
    Keys: SmallInt;
    WheelDelta: SmallInt;
    case Integer of
      0: (
        XPos: Smallint;
        YPos: Smallint);
      1: (
        Pos: TSmallPoint;
        Result: Longint);
  end;
{$ENDIF}


  TMsgTypeOption = (mtVScroll, mtHScroll, mtVHScroll,
                  mtVCursor, mtHCursor);
  TSendOption =  (soNone, soSendWM, soSendSim);

  PWheelCtrlRec = ^TWheelCtrlRec;
  TWheelCtrlRec = record
    Control: TWinControl;
    SendOpt: TSendOption;
    MsgType: TMsgTypeOption;
  end;

  PCtrlRecList = ^TCtrlRecList;
  TCtrlRecList = array[0..MaxListSize] of TWheelCtrlRec;

  TWheelMouse = class;

  TWheelCtrlList = class(TPersistent)
  private
    FVersion  : Integer;
    FFixupList: TStringList;
    FList     : PCtrlRecList;
    FCount    : Integer;
    FCapacity : Integer;
    FOwner    : TWheelMouse;
    procedure ReadData(Reader: TReader);
    procedure ReadOldData(Reader: TReader);
    procedure ReadVersion(Reader: TReader);
    procedure ExpandDataBlock;
    procedure FixupReferences;
    procedure WriteData(Writer: TWriter);
    procedure WriteVersion(Writer: TWriter);
  protected
    procedure DefineProperties(Filer: TFiler); override;
    function  GetCtrlRec(Index: Integer): TWheelCtrlRec;
    procedure SetCtrlRec(Index: Integer; Item: TWheelCtrlRec);
    procedure SetCapacity(NewCapacity: Integer);
  public
    constructor Create(AOwner: TWheelMouse);
    destructor Destroy; override;
    procedure Assign(Source: TPersistent); override;
    function  Add(AControl: TWinControl; MType: TMsgTypeOption;
                                       Opt: TSendOption): Integer;
    procedure Clear;
    procedure Delete(Index: Integer);
    function  Equals(WheelCtrlList: TWheelCtrlList): Boolean;
    function  IndexOf(AControl: TWinControl): Integer;
    property  Count: Integer read FCount;
    property  CtrlRecs[Index: Integer]: TWheelCtrlRec read GetCtrlRec
                        write SetCtrlRec; default;
  end;

  TWheelMsEvent = procedure(Sender: TObject; Shift: TShiftState;
                  X, Y, NumLines, PageOffSet: Integer) of object;

  TRawWheelMsEvent = procedure(zDelta, X, Y, ScrollLines: Integer) of object;

  TFakeMsgType = (fmVScroll, fmHScroll, fmUpDownArrows, fmLeftRightArrows);

{$IFDEF VERSION3_UP}
  TCrackerForm = class(TCustomForm);
{$ELSE}
  TCrackerForm = class(TForm);
{$ENDIF}

  TWheelMouse = class(TComponent)
  private
    { Private declarations }
    FAltSupport    : Boolean;
    FAutoMoveSBars : Boolean;
    FAutoUpdate    : Boolean;
    FSpecialCtrls  : TWheelCtrlList;
    FCurrentControl: TWinControl;
    FDeltaCount    : Integer;
    FMDISubclass   : Boolean;
    FMDIParentForm : TCrackerForm;
    FNativeSupport : Boolean;
    FNewWndProc    : Pointer;
    FOldWndProc    : Pointer;
    FParentWnd     : HWnd;
    FScrollLines   : Integer;
    FOnRawScroll   : TRawWheelMsEvent;
    FOnScroll      : TWheelMsEvent;
    FSendingWMMsg  : Boolean;
    FSendWMToAll   : Boolean;
    function  AdjustSubClassChain(var Message: TMessage): Boolean;
    procedure CalculateMovement(zDelta: Integer;
                       var PageOff, ScrUnit, ScrFactor: Integer);
    procedure SetupSubClass(AOwner: TComponent);
  protected
    { Protected declarations }
    procedure HandleWheelScroll(var Message: TMessage);
    procedure Loaded; override;
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
    procedure OwnerWndProc(var Message: TMessage);
    procedure SetNothingBool(Value: Boolean);
    procedure SetNothingInt(Value: Integer);
    procedure SetNewMouseInfo;
    procedure SetSpecialCtrls(Value: TWheelCtrlList);
    procedure UpdateMouseInfo;
  public
    { Public declarations }
    constructor Create(AOwner: TComponent); override;
    destructor  Destroy; override;
    function HandleVCLWheel(WheelDelta: Integer; MousePos: TPoint): Boolean;
    function HandleVCLWheelDown(MousePos: TPoint): Boolean;
    function HandleVCLWheelUp(MousePos: TPoint): Boolean;
  published
    { Published declarations }
    property AltSupport: Boolean read FAltSupport write SetNothingBool
                                                 stored False;
    property AutoMoveSBars: Boolean read FAutoMoveSBars
                                    write FAutoMoveSBars default True;
    property AutoUpdate: Boolean read FAutoUpdate write FAutoUpdate default True;
    property NativeSupport: Boolean read FNativeSupport write SetNothingBool
                                                 stored False;
    property OnRawScroll: TRawWheelMsEvent read FOnRawScroll write FOnRawScroll;
    property OnScroll: TWheelMsEvent read FOnScroll write FOnScroll;
    property ScrollLines: Integer read FScrollLines write SetNothingInt stored False;
    property SendWMToAll: Boolean read FSendWMToAll write FSendWMToAll default True;
    property SpecialCtrls: TWheelCtrlList read FSpecialCtrls write SetSpecialCtrls;
  end;

implementation

uses ZMouse {$IFDEF TRIAL}, TrialVer {$ENDIF} ;

const
  CurrentVer = 110;  { indicates data format version }

type
  TWMWheelMsRemove = packed record
    Msg: Cardinal;
    WndProc: Pointer;
    PrevWndProc: Pointer;
    Result: Longint;
  end;

  KeyData = record
    VirtKey : Byte;
    MouseKey: Byte;
  end;

var
  { message notifying of a TWheelMouse subclass removal }
  WM_WheelMouseRemove: Cardinal;

  KeyMap : array[0..4] of KeyData =
            ((VirtKey: VK_CONTROL; MouseKey: MK_CONTROL),
             (VirtKey: VK_LBUTTON; MouseKey: MK_LBUTTON),
             (VirtKey: VK_MBUTTON; MouseKey: MK_MBUTTON),
             (VirtKey: VK_RBUTTON; MouseKey: MK_RBUTTON),
             (VirtKey: VK_SHIFT; MouseKey: MK_SHIFT));

{ Local funcs/procs *************************************************}

procedure CheckForScrollBars(Control: TWinControl; var HScroll, VScroll: Boolean);
begin
  HScroll := GetWindowLong(Control.Handle, GWL_STYLE) and WS_HSCROLL <> 0;
  VScroll := GetWindowLong(Control.Handle, GWL_STYLE) and WS_VSCROLL <> 0;
end;

procedure ConvertMSHToWM(const MSHMsg: TMessage; var WMMsg: TWMMouseWheel);
var
  StateBuf: TKeyboardState;
  Index: Integer;
begin
  WMMsg := TWMMouseWheel(MSHMsg);
  with WMMsg do
  begin
     { in non native message all WPARAM is zDelta, in WM message
       zDelta = HIWORD(WPARAM) }
    WheelDelta := MSHMsg.WParam;
    Keys := 0;
    GetKeyboardState(StateBuf);
    for Index := Low(KeyMap) to High(KeyMap) do
      if StateBuf[KeyMap[Index].VirtKey] and $80 > 0 then
         Keys := Keys or KeyMap[Index].MouseKey;
  end;
end;

function DoKeyMessage(Control: TWinControl; VirtKey: Word;
                                     Count: Integer): Integer;
var
  I: Integer;
begin
  Result := 0;
  for I := 1 to Count do
  begin
    Result := SendMessage(Control.Handle, WM_KEYDOWN, VirtKey, 0);
    SendMessage(Control.Handle, WM_KEYUP, VirtKey, 0);
  end;
end;


{ TWheelCtrlList ****************************************************}

constructor TWheelCtrlList.Create(AOwner: TWheelMouse);
begin
  inherited Create;
  FOwner := AOwner;
  FVersion := CurrentVer;
end;

destructor TWheelCtrlList.Destroy;
begin
  Clear;
  inherited Destroy;
end;

function TWheelCtrlList.Add(AControl: TWinControl; MType: TMsgTypeOption;
                                   Opt: TSendOption): Integer;
var
  CtrlRec: TWheelCtrlRec;
begin
  with CtrlRec do
  begin
    Control := AControl;
    MsgType := MType;
    SendOpt := Opt;
  end;
  if FCount = FCapacity then ExpandDataBlock;
  if FCount < FCapacity then
  begin
    FList^[FCount] := CtrlRec;
    Result := FCount;
    Inc(FCount);
  end
  else Result := -1;
end;

procedure TWheelCtrlList.Assign(Source: TPersistent);
var
  CtrlList: TWheelCtrlList;
begin
  if Source is TWheelCtrlList then
  begin
    CtrlList := TWheelCtrlList(Source);
    Clear;
    SetCapacity(CtrlList.Count);
    System.Move(CtrlList.FList^, FList^, FCapacity * SizeOf(TWheelCtrlRec));
    FCount := CtrlList.Count;
  end
  else inherited Assign(Source);
end;

procedure TWheelCtrlList.Clear;
begin
  if FCount <> 0 then
  begin
    FCount := 0;
    SetCapacity(0);
  end;
end;

procedure TWheelCtrlList.DefineProperties(Filer: TFiler);

 function DoDataWrite: Boolean;
  begin
    if Filer.Ancestor <> nil then
    begin
      Result := True;
      if Filer.Ancestor is TWheelCtrlList then
        Result := not Equals(TWheelCtrlList(Filer.Ancestor));
    end
    else Result := Count > 0;
  end;

  function DoVersionWrite: Boolean;
  begin
    { Write version number if no ancestor and we are writing our control data
      or if we are writing data in a different format to an ancestor }
    Result := DoDataWrite;
    if Result then
      if (Filer.Ancestor <> nil) and (Filer.Ancestor is TWheelCtrlList) then
          Result := TWheelCtrlList(Filer.Ancestor).FVersion <> FVersion;
  end;

begin
  inherited DefineProperties(Filer);
  Filer.DefineProperty('Version', ReadVersion, WriteVersion, DoVersionWrite);
  Filer.DefineProperty('CtrlRecs', ReadData, WriteData, DoDataWrite);
  { SpecialCtrls for conversion of old data format only }
  Filer.DefineProperty('SpecialCtrls', ReadOldData, nil, False);
end;

procedure TWheelCtrlList.Delete(Index: Integer);
begin
  if (Index < 0) or (Index >= FCount) then
    raise EListError.Create('Invalid index to Delete procedure');
  Dec(FCount);
  if Index < FCount then
    System.Move(FList^[Index + 1], FList^[Index],
      (FCount - Index) * SizeOf(TWheelCtrlRec));
end;

procedure TWheelCtrlList.ExpandDataBlock;
var
  NewCapacity: Integer;
begin
  if FCapacity > 8 then NewCapacity := FCapacity + 16
    else if FCapacity > 4 then NewCapacity := FCapacity + 8
      else NewCapacity := FCapacity + 4;
  SetCapacity(NewCapacity);
end;

function TWheelCtrlList.Equals(WheelCtrlList: TWheelCtrlList): Boolean;
var
  I: Integer;
  CtrlRec1, CtrlRec2: TWheelCtrlRec;
begin
  Result := False;
  if FCount <> WheelCtrlList.Count then Exit;
  for I := 0 to FCount - 1 do
  begin
    CtrlRec1 := GetCtrlRec(I);
    CtrlRec2 := WheelCtrlList.GetCtrlRec(I);
    { compare control names (not references) and other data }
    if (CtrlRec1.Control.Name <> CtrlRec2.Control.Name) or
       (CtrlRec1.MsgType <> CtrlRec2.MsgType) or
       (CtrlRec1.SendOpt <> CtrlRec2.SendOpt) then Exit;
  end;
  Result := True;
end;

function TWheelCtrlList.GetCtrlRec(Index: Integer): TWheelCtrlRec;
begin
  Result := FList^[Index];
end;

function TWheelCtrlList.IndexOf(AControl: TWinControl): Integer;
begin
  for Result := 0 to FCount - 1 do
    if FList^[Result].Control = AControl then Exit;
  Result := -1;
end;

procedure TWheelCtrlList.FixupReferences;
var
  Index: Integer;
  Component: TComponent;
  CtrlRecPos: Integer;
begin
  if FFixupList = nil then Exit;
  for Index := FFixupList.Count - 1 downto 0 do
  begin
    CtrlRecPos := Integer(FFixupList.Objects[Index]);
    Component := TCrackerForm(FOwner.Owner).FindComponent(FFixupList[Index]);
    if Component <> nil then
    begin
      if Component is TWinControl then
        FList^[CtrlRecPos].Control := TWinControl(Component);
    end;
  end;
  FFixupList.Free;
 { important to set to nil as ReadData/ FixupReferences may be called again
   in form inheritance is used }
  FFixupList := nil;
  { check for any components not found in parent form and delete from list }
  if FCount > 0 then
    for Index := FCount - 1 to 0 do
      if FList^[Index].Control = nil then Delete(Index);
end;

procedure TWheelCtrlList.ReadData(Reader: TReader);
var
  CtrlRec: TWheelCtrlRec;
  Index: Integer;
begin
  Clear;
  Reader.ReadListBegin;
  while not Reader.EndOfList do
  begin
    if FFixupList = nil then
      FFixupList := TStringList.Create;
    Index := FFixupList.Add(Reader.ReadString);
    CtrlRec.SendOpt := TSendOption(Reader.ReadInteger);
    CtrlRec.MsgType := TMsgTypeOption(Reader.ReadInteger);
    FFixupList.Objects[Index] :=
      Pointer(Add(nil, CtrlRec.MsgType, CtrlRec.SendOpt));
  end;
  Reader.ReadListEnd;
  { ReadData can be called to update components properties when using
    form inheritance. In that event 'Loaded' will not be called so call
    FixupReferences from here }
  if FOwner.ComponentState * [csUpdating, csAncestor] =
      [csUpdating, csAncestor] then FixupReferences;
  FVersion := CurrentVer;
end;


procedure TWheelCtrlList.ReadOldData(Reader: TReader);
{ conversion routine for use when reading older TWheelMouse data }
type
  TOldTypeOption = (otNone, otVScroll, otHScroll, otVHScroll,
                  otVCursor, otHCursor);
  TGenOption =  (goSendWM, goSendSim);
  TGenOptions = set of TGenOption;
  TOldCtrlRec = packed record
    MsgType:  TOldTypeOption;
    GenOpt : TGenOptions;
    Spare: Word;                 { Padding to require 4 byte size }
  end;

var
  CtrlRec: TWheelCtrlRec;
  Index: Integer;
  OldCtrlRec: TOldCtrlRec;
begin
  Clear;
  Reader.ReadListBegin;
  while not Reader.EndOfList do
  begin
    if FFixupList = nil then
      FFixupList := TStringList.Create;
    Index := FFixupList.Add(Reader.ReadString);
    OldCtrlRec := TOldCtrlRec(Integer(Reader.ReadInteger));
    if OldCtrlRec.MsgType = otNone then OldCtrlRec.MsgType := otVScroll;
    CtrlRec.MsgType := TMsgTypeOption(Pred(OldCtrlRec.MsgType));
    if goSendSim in OldCtrlRec.GenOpt then CtrlRec.SendOpt := soSendSim
    else if goSendWM in OldCtrlRec.GenOpt then CtrlRec.SendOpt := soSendWM;
    FFixupList.Objects[Index] :=
      Pointer(Add(nil, CtrlRec.MsgType, CtrlRec.SendOpt));
  end;
  Reader.ReadListEnd;
  { ReadData can be called to update components properties when using
    form inheritance. In that event 'Loaded' will not be called so call
    FixupReferences from here }
  if FOwner.ComponentState * [csUpdating, csAncestor] =
      [csUpdating, csAncestor] then FixupReferences;
  FVersion := CurrentVer;
end;

procedure TWheelCtrlList.ReadVersion(Reader: TReader);
begin
  FVersion := Reader.ReadInteger;
end;

procedure TWheelCtrlList.SetCapacity(NewCapacity: Integer);
begin
  ReallocMem(FList, NewCapacity * SizeOf(TWheelCtrlRec));
  FCapacity := NewCapacity;
end;

procedure TWheelCtrlList.SetCtrlRec(Index: Integer; Item: TWheelCtrlRec);
begin
  Item.Control := FList^[Index].Control; // Prevent control changes
  FList^[Index] := Item;
end;

procedure TWheelCtrlList.WriteData(Writer: TWriter);
var
  Index : Integer;
begin
  Writer.WriteListBegin;
  for Index := 0 to Count - 1 do
  begin
    Writer.WriteString(FList^[Index].Control.Name);
    Writer.WriteInteger(Integer(FList^[Index].SendOpt));
    Writer.WriteInteger(Integer(FList^[Index].MsgType));
  end;
  Writer.WriteListEnd;
end;

procedure TWheelCtrlList.WriteVersion(Writer: TWriter);
begin
  Writer.WriteInteger(FVersion);
end;


{ TWheelMouse ******************************************************}

constructor TWheelMouse.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
 {$IFDEF TRIAL} TrialCheck; {$ENDIF}
  FSendWMToAll := True;
  FAutoMoveSBars := True;
  FAutoUpdate := True;
  FSpecialCtrls := TWheelCtrlList.Create(Self);
  SetNewMouseInfo;
{$IFDEF VERSION2_DOWN}
  if not (AOwner is TForm) then
    raise Exception.Create('Owner must be a TForm descendant');
{$ELSE}  { Delphi 3 + }
   if not (AOwner is TCustomForm) then
      raise Exception.Create('Owner must be a TCustomForm descendant');
{$ENDIF}
   { if designing then don't set up subclass }
   if not (csDesigning in ComponentState) then SetupSubclass(AOwner);
end;

destructor TWheelMouse.Destroy;
begin
  FSpecialCtrls.Free;

  if not (csDesigning in ComponentState) then
  begin
  { Restore subclassed form's window proc if it still points to us }
    if Pointer(GetWindowLong(FParentWnd, GWL_WNDPROC)) = FNewWndProc then
      SetWindowLong(FParentWnd, GWL_WNDPROC, Longint(FOldWndProc))

    else
     { Notify other TWheelMouse comps. that are sub-classing the
      same window of this removal }
      SendMessage(FParentWnd, WM_WheelMouseRemove,
              Integer(FNewWndProc), Integer(FOldWndProc));

    if FNewWndProc <> nil then FreeObjectInstance(FNewWndProc);
  end;
  inherited Destroy;
end;


function TWheelMouse.AdjustSubClassChain(var Message: TMessage): Boolean;
begin
  { if the TWheelMouse being removed is the previous subclass then
    adjust our FOldWndProc link }
  Result := FOldWndProc = TWMWheelMsRemove(Message).WndProc;
  if Result then
    FOldWndProc := TWMWheelMsRemove(Message).PrevWndProc;
end;

procedure TWheelMouse.CalculateMovement(zDelta: Integer;
                       var PageOff, ScrUnit, ScrFactor: Integer);
var
  NewDelta : Integer;
begin
  NewDelta := FDeltaCount + zDelta;
  ScrUnit := NewDelta div WHEEL_DELTA;
  FDeltaCount := NewDelta mod WHEEL_DELTA;
  PageOff := 0;
  if FScrollLines = Integer(WHEEL_PAGESCROLL) then
  begin
     PageOff := ScrUnit;
     ScrUnit := 0;
     ScrFactor := 1;
  end
  else ScrFactor := FScrollLines;
end;

function TWheelMouse.HandleVCLWheel(WheelDelta: Integer;
                                          MousePos: TPoint): Boolean;
var
  Message: TMessage;
begin
  Result := False;
 { if currently sending WM_MOUSEWHEEL message then must be on Win95 and
   focused control handling is set to send WM_MOUSEWHEEL message or the
   SendWMToAll is property is set. If this is the case and we either get
   here via the control's or one of its parents wheel events, hence no
   extra handling is required. }
  if FSendingWMMsg then Exit;
  with Message do
  begin
    Msg := Msh_MsgWheelMove;
    WParam := WheelDelta;
    LParam := DWORD(PointToSmallPoint(MousePos));
  end;
  HandleWheelScroll(Message);
  Result := Message.Result = 0;
end;

function TWheelMouse.HandleVCLWheelDown(MousePos: TPoint): Boolean;
begin
  Result := HandleVCLWheel(-WHEEL_DELTA, MousePos)
end;

function TWheelMouse.HandleVCLWheelUp(MousePos: TPoint): Boolean;
begin
  Result := HandleVCLWheel(WHEEL_DELTA, MousePos);
end;

var
  ScrollBarMsg: array[False..True] of Integer = (WM_VSCROLL, WM_HSCROLL);
  ScrollBarAction: array[False..True, False..True] of Integer =
                 ((SB_LINEDOWN, SB_LINEUP), (SB_PAGEDOWN, SB_PAGEUP));
  CursorKey: array[False..True, False..True] of Integer =
                ((VK_DOWN, VK_UP), (VK_LEFT, VK_RIGHT));

procedure TWheelMouse.HandleWheelScroll(var Message: TMessage);
var
  ScrollUnit,
  ScrollFactor: Integer;
  PageOffset  : Integer;
  Index       : Integer;
  FakeScroll  : Integer;
  FakeMsg     : Integer;
  MouseMsg    : TWMMouseWheel;
  CtrlRec     : TWheelCtrlRec;
  HasVScroll,
  HasHScroll  : Boolean;
  LocalPos    : TPoint;
begin
  with TCrackerForm(Owner) do
    if ActiveControl <> FCurrentControl then
    begin
      FCurrentControl := ActiveControl;
      FDeltaCount := 0; { reset delta counter if on a new control }
    end;
  if FCurrentControl = nil then Exit; { if no active control then exit }

   { if not WM_MOUSEWHEEL messages then convert message }
  if Message.Msg <> WM_MOUSEWHEEL then ConvertMSHToWM(Message, MouseMsg)
  else MouseMsg := TWMMouseWheel(Message);

  CheckForScrollBars(FCurrentControl, HasHScroll, HasVScroll);
  Index := FSpecialCtrls.IndexOf(FCurrentControl);
  if Index <> -1 then CtrlRec := FSpecialCtrls.CtrlRecs[Index]
  else { if no defined action for a control, check for general actions }
    begin
      CtrlRec.SendOpt := soNone;  { set default 'action' }
      if FAutoMoveSBars and (HasVScroll or HasHScroll) then
      begin
        CtrlRec.SendOpt := soSendSim;
        CtrlRec.MsgType := mtVHScroll;
      end
      else
        if FSendWMToAll then CtrlRec.SendOpt := soSendWM;
    end;

  if CtrlRec.SendOpt = soNone then Message.Result := 1   { unhandled }
  else Message.Result := 0;  { handled }

  with MouseMsg do
  begin
    CalculateMovement(WheelDelta, PageOffSet, ScrollUnit, ScrollFactor);
    if Assigned(FOnScroll) then
    begin
      LocalPos := FCurrentControl.ScreenToClient(SmallPointToPoint(Pos));
      FOnScroll(FCurrentControl, KeysToShiftState(Keys),
             LocalPos.X, LocalPos.Y, ScrollUnit * ScrollFactor, PageOffSet);
    end;

    if Assigned(FOnRawScroll) then
      FOnRawScroll(WheelDelta, XPos, YPos, FScrollLines);

    if (CtrlRec.SendOpt = soSendWM) and not FNativeSupport then
    begin
      FSendingWMMsg := True;  // prevent recursion
      SendMessage(FCurrentControl.Handle, WM_MOUSEWHEEL,
              MAKEWPARAM(Keys, WheelDelta), MAKELPARAM(XPos, YPos));
      FSendingWMMsg := False;
    end;
    { once events are done, set ScrollUnit to PageOffset if the system is set
      for page scrolling, otherwise handling will not complete correctly }
    if PageOffset <> 0 then ScrollUnit := PageOffSet;
    if (CtrlRec.SendOpt = soSendSim) and (FCurrentControl.Enabled) and
      (FCurrentControl.Visible) then
      case CtrlRec.MsgType of
        mtVCursor, mtHCursor:
          DoKeyMessage(FCurrentControl,
                       CursorKey[CtrlRec.MsgType = mtHCursor, ScrollUnit > 0],
                       Abs(ScrollUnit));
        mtVScroll, mtHScroll, mtVHScroll:
        begin
          if CtrlRec.MsgType = mtVHScroll then
            { do vert scroll unless no vert bar and has a horiz bar }
             FakeMsg := ScrollBarMsg[not HasVScroll]
          else FakeMsg := ScrollBarMsg[CtrlRec.MsgType = mtHScroll];
          { don't send scroll bar message unless it has the right type }
          if ((FakeMsg = WM_VSCROLL) and HasVScroll) or
             ((FakeMsg = WM_HSCROLL) and HasHScroll) then
          begin
            FakeScroll := ScrollBarAction[PageOffset <> 0, ScrollUnit > 0];
              for Index := 0 to Abs(ScrollUnit * ScrollFactor) - 1 do
                 SendMessage(FCurrentControl.Handle, FakeMsg, FakeScroll, 0)
          end;
        end;
      end; {case}
  end; { with MouseMsg}
end;

procedure TWheelMouse.Loaded;
begin
  inherited Loaded; { Clears csLoading flag }
  FSpecialCtrls.FixupReferences;
end;

procedure TWheelMouse.Notification(AComponent: TComponent; Operation: TOperation);
var
  Index: Integer;
begin
  inherited Notification(AComponent, Operation);
  { check for deletion of a control in our list }
  if (Operation = opRemove) and not (csDestroying in ComponentState) and
       (AComponent is TWinControl) then
  begin
    Index := FSpecialCtrls.IndexOf(TWinControl(AComponent));
    if Index <> -1 then FSpecialCtrls.Delete(Index);
  end;
end;

procedure TWheelMouse.OwnerWndProc(var Message: TMessage);
begin
  { Note: WM_MOUSEWHEEL messages only get here if they are unhandled by the
          focused control }
  with Message do
  begin
    if (Msg = Cardinal(Msh_MsgWheelMove)) or
       ((Msg = WM_MOUSEWHEEL) and not FSendingWMMsg) then
    begin
     { WM_MOUSEWHEEL messages can be sent to an inactive window. Processing
       these for VCL controls can cause the Active property of the application
       and form to be set to Active even though the app is not focused.
       Therefore these are ignored by TWheelMouse }
     {$IFDEF VERSION2_DOWN}
         if Application.Active and (Screen.ActiveForm = Owner) then
     {$ELSE}
         if Application.Active and (Screen.ActiveCustomForm = Owner) then
     {$ENDIF}
        begin
          HandleWheelScroll(Message);
          if Message.Result <> 0 then
            Result := CallWindowProc(FOldWndProc, FParentWnd, Msg, wParam, lParam);
        end
        else Result := CallWindowProc(FOldWndProc, FParentWnd, Msg, wParam, lParam);
    end
    else
      begin
       { use this proc hook to also get notification of settings changes }
        if (Msg = WM_SETTINGCHANGE) and FAutoUpdate then UpdateMouseInfo;
        { only one sub-class can need adjusting so stop further handling if
          it is us }
        if (Msg <> WM_WheelMouseRemove) or not AdjustSubClassChain(Message) then
          Result := CallWindowProc(FOldWndProc, FParentWnd, Msg, wParam, lParam);
      end;
  end;
end;

procedure TWheelMouse.SetNothingBool(Value: Boolean);
begin
  { Do nothing }
end;

procedure TWheelMouse.SetNothingInt(Value: Integer);
begin
  { Do nothing }
end;

procedure TWheelMouse.SetNewMouseInfo;
begin
  FScrollLines := MsWheelScrollLines;
  FNativeSupport := MsWheelNativeSupp;
  FAltSupport := MsWheelAltSupp;
end;

procedure TWheelMouse.SetSpecialCtrls(Value: TWheelCtrlList);
begin
  FSpecialCtrls.Assign(Value);
end;

procedure TWheelMouse.SetupSubclass(AOwner: TComponent);
var
  MDIParent: HWnd;
  Index: Integer;
begin
{$IFDEF VERSION2_DOWN}
  { need to subclass MDI parent if on an MDI child form and the OS does
    not have native support }
  FMDISubClass := (TForm(AOwner).FormStyle = fsMDIChild) and
                  not MsWheelNativeSupp;
  if FMDISubClass then
  begin
    MDIParent := GetParent(GetParent(TForm(AOwner).Handle));
    with Screen do
      for Index := 0 to FormCount - 1 do
        if Forms[Index].Handle = MDIParent then
        begin
          FMDIParentForm := TCrackerForm(Forms[Index]);
          Break;
        end;
    if FMDIParentForm = nil then
      raise Exception.Create('Cannot locate MDI parent form');
    FParentWnd := FMDIParentForm.Handle;
  end
  else FParentWnd := TForm(AOwner).Handle;

{$ELSE}  { Delphi 3 + }
   { need to subclass MDI parent if on an MDI child form and the OS does
    not have native support }
  FMDISubClass := (TCrackerForm(AOwner).FormStyle = fsMDIChild) and
                  not MsWheelNativeSupp;
  if FMDISubClass then
  begin
    MDIParent := GetParent(GetParent(TCustomForm(AOwner).Handle));
    with Screen do
      for Index := 0 to CustomFormCount - 1 do
        if CustomForms[Index].Handle = MDIParent then
        begin
          FMDIParentForm := TCrackerForm(CustomForms[Index]);
          Break;
        end;
    if FMDIParentForm = nil then
      raise Exception.Create('Cannot locate MDI parent form');
    FParentWnd := FMDIParentForm.Handle;
  end
  else FParentWnd := TCustomForm(AOwner).Handle;
{$ENDIF}

  { create new wnd proc to be used by the owning form }
  FNewWndProc := MakeObjectInstance(OwnerWndProc);
  { replace form's wnd proc with the new wnd proc }
  FOldWndProc := Pointer(SetWindowLong(FParentWnd, GWL_WNDPROC,
                              Longint(FNewWndProc)));
  { raise exception if SetWindowLong didn't work }
  if FOldWndProc = nil then
    raise Exception.Create('Unable to subclass owner''s WndProc');
end;

procedure TWheelMouse.UpdateMouseInfo;
begin
  UpdateWheelMouseInfo;
  SetNewMouseInfo;
end;

initialization
  WM_WheelMouseRemove :=
    RegisterWindowMessage('TWheelMouse removal notification');
{$IFDEF TRIAL}
  TrialCheck;
{$ENDIF}
end.


